local L = LibStub("AceLocale-3.0"):GetLocale("AffDotsDruid")

-- Damage multiplier for the various buffs
local const = {
    ["tf"]           = 1.15, -- Tigers Fury
    ["sr"]           = 1.4,  -- Savage Roar
    ["doc"]          = 1.3,  -- Dream of Cenarius    
    ["sb"]           = 1.2,  -- Skullbanner
    ["meta"]         = 1.06  -- Meta
}

-- Spell & Aura ids
local spell_id = {
    ["rip"]           = 1079,
    ["rake"]          = 1822,
    ["thrash"]        = 106830,
    ["shred"]         = 5221,
    ["ravage"]        = 6785,
    ["mangle"]        = 33876,
    ["pounce"]        = 33876,
    ["swipe"]         = 33876,
    ["maim"]          = 33876,
    
    ["fb"]            = 22568,
    ["tf"]            = 5217,
    ["sr"]            = 127538,
    ["doc"]           = 145152,
    
    ["sb"]            = 114207, -- Skullbanner
}

local options
local playerGuid, spec
local playerLevel = UnitLevel("player")
local ripDps, rakeDps, thrashDps = 0, 0, 0
local critBonus = 1
local timer, throttle       = 0, 0.1
local dot_damage, targets   = AffDots.dot_damage, AffDots.targets
dot_damage[spell_id.rip] = {}
dot_damage[spell_id.rake] = {}
dot_damage[spell_id.thrash] = {}
local tf, sr, doc, sb = 1, 1, 1, 1
local critDepr = false
local notSbScan = true
local oldCp = 0
local colorSwitch = 1
local protect

-- Save frequently used globals as locals for performance reasons
local GetMasteryEffect = GetMasteryEffect
local GetCritChance    = GetCritChance
local UnitAttackPower  = UnitAttackPower
local GetComboPoints   = GetComboPoints
local GetTime          = GetTime
local UnitLevel        = UnitLevel
local UnitAura         = UnitAura
local UnitGUID         = UnitGUID
local UnitDebuff       = UnitDebuff
local math_floor       = math.floor
local math_abs         = math.abs
local max              = max
local pairs            = pairs

local AffDots          = AffDots
local TidyPlates       = TidyPlates


local getCritMultiplier 
local _, _, rake_icon    = GetSpellInfo(spell_id.rake)
local _, _, rip_icon     = GetSpellInfo(spell_id.rip)
local _, _, thrash_icon  = GetSpellInfo(spell_id.thrash)
local doc_name = GetSpellInfo(spell_id.doc)
local tf_name  = GetSpellInfo(spell_id.tf)
local sr_name  = GetSpellInfo(spell_id.sr)
local sb_name  = GetSpellInfo(spell_id.sb)

local function Round(num) return math_floor(num+.5) end
local function SecondsRound(num)
    if num > 2 then return math_floor(num+.5)
    else return math_floor(num*10+.5)/10 end
end

-- Rip damage formula
local function calcRipDpsDisplay(ap, tf, sr, doc, mastery, crit, cp)
    local damage
    if cp == 0 then 
        damage = 0 
    else
        -- Dot dmg * tick count * dmg buffs
        damage = 8*(113+(320 + 0.0484*ap)*cp) * (crit*mastery*sr*tf*doc)
    end
    -- 1600 = duration * 100
    return Round(damage/1600)/10
end

-- Rake damage formula
local function calcRakeDpsDisplay(ap, tf, sr, doc, mastery, crit)
    -- dot ticks only
    local damage = 5*(99+0.3*ap) * (crit*mastery*sr*tf*doc)
    return Round(damage/1500)/10
end

-- Thrash damage formula
local function calcThrashDpsDisplay(ap, tf, sr, doc, mastery, crit)
    -- dot ticks only
    local damage = 5*(686 + ap*0.141) * (crit*mastery*sr*tf*doc)
    return Round(damage/1500)/10
end

local function getMasteryMultiplier()
    local m = GetMasteryEffect()
    m = 1 + m/100
    return m
end

-- Calculates Crit damage multiplier with Crit depression
local function getCMwCDepr(bonus, unit)
    local critDepression
    local unitLevel = UnitLevel(unit)
    -- Crit depression in MoP
    -- http://maintankadin.failsafedesign.com/forum/viewtopic.php?f=4&t=33304#p755100
    -- http://blue.mmo-champion.com/topic/227466-beta-class-balance-analysis/#post97
    if unitLevel == -1 then
        critDepression = 3
    else
        critDepression = max(unitLevel - playerLevel, 0)
    end
    
    local c = (GetCritChance() - critDepression) / 100
    if c > 1 then c = 1 end
    c = 1 + c*bonus
    return c
end

-- Calculates Crit damage multiplier without Crit depression
local function getCM(bonus, unit)
    local c = GetCritChance() / 100
    if c > 1 then c = 1 end
    c = 1 + c*bonus
    return c
end

local function getDamageStats(unit) 
    unit = unit or "target"
    local base, posBuff, negBuff = UnitAttackPower("player")
    local ap      = base + posBuff + negBuff
    local mastery = getMasteryMultiplier() 
    local crit    = getCritMultiplier(critBonus * sb, unit)
    local cp      = GetComboPoints("player", unit)
    if cp ~= 0 then
        oldCp = cp
    end
    return ap, tf, sr, doc, mastery, crit, cp
end

local function onUpdate(self,elapsed)
    timer = timer + elapsed;
    if timer >= throttle then
		for k,v in pairs(AffDots.track) do
			AffDots.track[k].update(AffDots.track[k])
			AffDots:FrameDraw(AffDots.track[k])
		end
		
		if options.tidyplates.enable and TidyPlates then 
			AffDots:Tidy() 
		end
		--[[
		if options.burst.enable then 
			AffDots.Update.burst() 
		end
        ]]
        timer = 0
    end
end

AffDots.Update = {
    dot = function(f)
        local guid = UnitGUID(f.target)
        local _,_,_,_,_,duration,expires = UnitDebuff(f.target,f.name,f.rank,"player")
        local datanew = f.datanew
        if duration and guid and targets[guid..f.spell] then
            if(math_abs(expires-f.timer)>1) then
                datanew.expires = expires
                datanew.duration = duration
                datanew.setcd = true
                f.timer = expires
            end
            datanew.color = options["color"..AffDots:FindColor(guid,expires - GetTime(),f.spell,f.pandemic)]
            datanew.tick = targets[guid..f.spell][3]
            datanew.t2 = SecondsRound(f.timer-GetTime())
            if options.reltext then
                datanew.t1 = Round(dot_damage[f.spell][1]*100/targets[guid..f.spell][1])
            elseif options.layout == "bars" then
                datanew.t1 = dot_damage[f.spell][1].."/"..targets[guid..f.spell][1]
            else
                datanew.t1 = dot_damage[f.spell][1].."\n"..targets[guid..f.spell][1]
            end
        else
            datanew.color = options.color7
            datanew.t1 = 0
            datanew.t2 = ""
            datanew.tick = 0
            timer = 0
            datanew.hidecd = true
        end
    end,
}

local GetDotDmg = {
    function()
    end,
    function(self, event, unit, ...)
        -- from all tracked events the first parameter is either unit or nil
        if unit and unit ~= "player" then return end
        local ap, tf, sr, doc, mastery, crit, cp = getDamageStats()
        -- Rip
        dot_damage[spell_id.rip][1]   = calcRipDpsDisplay(ap, tf, sr, doc, mastery, crit, cp)
        dot_damage[spell_id.rip][2]   = 2
        -- Rake
        dot_damage[spell_id.rake][1]   = calcRakeDpsDisplay(ap, tf, sr, doc, mastery, crit)
        dot_damage[spell_id.rake][2]   = 3
        -- Thrash
        dot_damage[spell_id.thrash][1] = calcThrashDpsDisplay(ap, tf, sr, doc, mastery, crit)
        dot_damage[spell_id.thrash][2] = 3
        
    end,
    function() end,
    function() end,
}

local CombatLog = {
    function() end,
    function(_,_,_,event_type,_,source_GUID,_,_,_,dest_GUID,_,_,_, ...)
        if source_GUID ~= playerGuid and (notSbScan or dest_GUID ~= playerGuid) then return end
        local spell = ...

        if event_type == "SPELL_AURA_APPLIED" then
            if spell == spell_id.rip then
                targets[dest_GUID..spell_id.rip]    = targets[dest_GUID..spell_id.rip] or {}
                targets[dest_GUID..spell_id.rip][1] = ripDps
                targets[dest_GUID..spell_id.rip][2] = GetTime()
                targets[dest_GUID..spell_id.rip][3] = dot_damage[spell_id.rip][2]
            elseif spell == spell_id.rake then
                targets[dest_GUID..spell_id.rake]    = targets[dest_GUID..spell_id.rake] or {}
                targets[dest_GUID..spell_id.rake][1] = rakeDps
                targets[dest_GUID..spell_id.rake][2] = GetTime()
                targets[dest_GUID..spell_id.rake][3] = dot_damage[spell_id.rake][2]
            elseif spell == spell_id.thrash then
                targets[dest_GUID..spell_id.thrash]    = targets[dest_GUID..spell_id.thrash] or {}
                targets[dest_GUID..spell_id.thrash][1] = thrashDps
                targets[dest_GUID..spell_id.thrash][2] = GetTime()
                targets[dest_GUID..spell_id.thrash][3] = dot_damage[spell_id.thrash][2]
            elseif spell == spell_id.tf then
                tf = const.tf
            elseif spell == spell_id.sr then
                sr = const.sr
            elseif spell == spell_id.doc then
                doc = const.doc
            elseif spell == spell_id.sb then
                sb = const.sb
            end            
        elseif event_type == "SPELL_AURA_REFRESH" then
            if spell == spell_id.rip and GetTime() - protect > 0.5 then
                targets[dest_GUID..spell_id.rip]    = targets[dest_GUID..spell_id.rip] or {}
                targets[dest_GUID..spell_id.rip][1] = ripDps
                targets[dest_GUID..spell_id.rip][2] = GetTime()
                targets[dest_GUID..spell_id.rip][3] = dot_damage[spell_id.rip][2]
            elseif spell == spell_id.rake then
                targets[dest_GUID..spell_id.rake]    = targets[dest_GUID..spell_id.rake] or {}
                targets[dest_GUID..spell_id.rake][1] = rakeDps
                targets[dest_GUID..spell_id.rake][2] = GetTime()
                targets[dest_GUID..spell_id.rake][3] = dot_damage[spell_id.rake][2]
            elseif spell == spell_id.thrash then
                targets[dest_GUID..spell_id.thrash]    = targets[dest_GUID..spell_id.thrash] or {}
                targets[dest_GUID..spell_id.thrash][1] = thrashDps
                targets[dest_GUID..spell_id.thrash][2] = GetTime()
                targets[dest_GUID..spell_id.thrash][3] = dot_damage[spell_id.thrash][2]
            end
        elseif event_type == "SPELL_DAMAGE" then
            if (spell == spell_id.shred or spell == spell_id.mangle or spell == spell_id.fb or spell == spell_id.ravage) then
                protect = GetTime()
            end
        elseif event_type == "SPELL_CAST_SUCCESS" then
            -- Snapshot dps here, because dmg ingame seems to be calculated at SPELL_CAST_SUCCESS rather than SPELL_AURA_*
            -- DoC buff is gone on SPELL_AURA_* events anyway
            if spell == spell_id.rip then
                -- try to fix t16 bug
                local ap, tf, sr, doc, mastery, crit, cp = getDamageStats()
                ripDps = calcRipDpsDisplay(ap, tf, sr, doc, mastery, crit, oldCp)
            elseif spell == spell_id.rake then
                rakeDps = calcRakeDpsDisplay(getDamageStats())
            elseif spell == spell_id.thrash then
                thrashDps = calcThrashDpsDisplay(getDamageStats())
            end
        elseif event_type == "SPELL_AURA_REMOVED" then
            if spell == spell_id.tf then
                tf = 1
            elseif spell == spell_id.sr then
                sr = 1
            elseif spell == spell_id.doc then
                doc = 1
            elseif spell == spell_id.sb then
                sb = 1
            end 
        end
    end,
    function() end,
    function() end,
}

function AffDots:InitSpec()
    AffDots:ReleaseFrames()
    wipe(AffDots.track)
    if spec == 2 then
        AffDots.execute_percent = 25
        AffDots:Track( spell_id.rip, AffDots.Update.dot, "target",   16,  2, options.feral.rip_order,  options.feral.show_rip)
        AffDots:Track( spell_id.rake, AffDots.Update.dot, "target",   15,  3, options.feral.rake_order,  options.feral.show_rake)
        AffDots:Track( spell_id.thrash, AffDots.Update.dot, "target",   16,  3, options.feral.thrash_order,  options.feral.show_thrash)
        if options.focus then
            AffDots:Track( spell_id.rip, AffDots.Update.dot,  "focus",    16,  2, options.feral.rip_order,  options.feral.show_rip)
            AffDots:Track( spell_id.rake, AffDots.Update.dot,  "focus",   15,  3, options.feral.rake_order,  options.feral.show_rake)
            AffDots:Track( spell_id.thrash, AffDots.Update.dot, "focus",   16,  3, options.feral.thrash_order,  options.feral.show_thrash)
        end
		AffDots.tidyspells = {
            [rake_icon] = {spell = spell_id.rake, pandemic = 0},
            [rip_icon] = {spell = spell_id.rip, pandemic = 0},
            [thrash_icon] = {spell = spell_id.thrash, pandemic = 0},
        }
    end
    critBonus = options.feral.meta and const.meta or 1
    getCritMultiplier = options.feral.crit_depr and getCMwCDepr or getCM
    notSbScan = not options.feral.sb_scan
    colorSwitch = options.feral.color_switch

    AffDots:InitFrames()
end

function AffDots:InitializeClass()
    playerGuid = UnitGUID("player")
    AffDots:RegisterEvent("PLAYER_TALENT_UPDATE", "InitializeClass")
    AffDots:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    AffDots:UnregisterEvent("COMBAT_RATING_UPDATE")
    AffDots:UnregisterEvent("SPELL_POWER_CHANGED")
    AffDots:UnregisterEvent("PLAYER_DAMAGE_DONE_MODS")
    AffDots:UnregisterEvent("UNIT_STATS")
    AffDots:UnregisterEvent("UNIT_ATTACK_POWER")
    AffDots:UnregisterEvent("UNIT_COMBO_POINTS")
    AffDots:UnregisterEvent("UNIT_AURA")
    AffDots:UnregisterEvent("PLAYER_REGEN_ENABLED")
    AffDots:UnregisterEvent("PLAYER_REGEN_DISABLED")
    AffDots:UnregisterEvent("PLAYER_FOCUS_CHANGED")
    AffDots:UnregisterEvent("MODIFIER_STATE_CHANGED")
    AffDotsTarget:SetScript("OnUpdate", nil)
    AffDotsTarget:Hide(); AffDotsFocus:Hide(); AffDotsBurst:Hide()
    spec = GetSpecialization() or ""
    if (spec == 2) then
        options = self.db.profile
        options.feral = options.feral or {}
        options.feral.rip_order = options.feral.rip_order or 0
        options.feral.rake_order = options.feral.rake_order or 1
        options.feral.thrash_order = options.feral.thrash_order or 2
        options.feral.color_switch = options.feral.color_switch or 1
        options.feral.show_rip = options.feral.show_rip == nil or options.feral.show_rip
        options.feral.show_rake = options.feral.show_rake == nil or options.feral.show_rake
        options.feral.show_thrash = options.feral.show_thrash == nil or options.feral.show_thrash
        options.feral.crit_depr = options.feral.crit_depr or false
        options.feral.meta = options.feral.meta or false
        options.feral.sb_scan = options.feral.sb_scan or false

        critBonus = options.feral.meta and const.meta or 1
        getCritMultiplier = options.feral.crit_depr and getCMwCDepr or getCM
        notSbScan = not options.feral.sb_scan
        colorSwitch = options.feral.color_switch
        
        AffDots.GetDotDmg = GetDotDmg[spec]
        AffDots.CombatLog = CombatLog[spec]
        AffDots:GetDotDmg()
        AffDots:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "CombatLog")
        AffDots:RegisterEvent("COMBAT_RATING_UPDATE", "GetDotDmg")
        AffDots:RegisterEvent("SPELL_POWER_CHANGED", "GetDotDmg")
        AffDots:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "GetDotDmg")
        AffDots:RegisterEvent("UNIT_STATS", "GetDotDmg") 
        AffDots:RegisterEvent("UNIT_ATTACK_POWER", "GetDotDmg") 
        AffDots:RegisterEvent("UNIT_COMBO_POINTS", "GetDotDmg") 
        AffDots:RegisterEvent("UNIT_AURA", "GetDotDmg") 
        AffDots:RegisterEvent("PLAYER_REGEN_ENABLED")  
        AffDots:RegisterEvent("PLAYER_REGEN_DISABLED")
        AffDots:RegisterEvent("PLAYER_FOCUS_CHANGED")
        AffDots:RegisterEvent("MODIFIER_STATE_CHANGED")
        AffDotsTarget:SetScript("OnUpdate", onUpdate)
        AffDots:InitSpec()
    end
end

function AffDots:GetSpecOptions(spec)
    if spec == "DRUID2" then
        self.options.args.burst = nil
        self.options.args.spec.name = L["Feral"]
        self.options.args.spec.args = {
			show_rip = {
				order = 1, name = L["Show Rip"], type = "toggle",
				set = function(info,val) options.feral.show_rip = not options.feral.show_rip; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_rip end
			},
			show_rake = {
				order = 2, name = L["Show Rake"], type = "toggle",
				set = function(info,val) options.feral.show_rake = not options.feral.show_rake; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_rake end
			},
			show_thrash = {
				order = 3,name = L["Show Thrash"],type = "toggle",
				set = function(info,val) options.feral.show_thrash = not options.feral.show_thrash; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_thrash end
			},
            desc1 = {
                order = 4,name = L["Order:"],type = "description",fontSize="medium"
            },
			rip_order = {
				order = 5,name = L["Rip"],type = "range",
				min=1, max=3, step=1, softMin=1, softMax=3,
				set = function(info,val) options.feral.rip_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.rip_order + 1 end
			},
			rake_order = {
				order = 6,name = L["Rake"],type = "range",
				min=1, max=3, step=1, softMin=1, softMax=3,
				set = function(info,val) options.feral.rake_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.rake_order + 1 end
			},
			thrash_order = {
				order = 6, name = L["Thrash"], type = "range",
				min=1, max=3, step=1, softMin=1, softMax=3,
				set = function(info,val) options.feral.thrash_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.thrash_order + 1 end
			},
            desc2 = {
                order = 7, name = L["Advanced options:"], type = "description", fontSize="medium"
            },
			crit_depr = {
				order = 8,name = L["Use crit depression"], type = "toggle",
                desc = L["-1% crit for every level; -3% for bosses\nMainly interesting if you are not showing relative numbers"],
				set = function(info,val) options.feral.crit_depr = not options.feral.crit_depr; AffDots:InitSpec() end,
				get = function(info) return options.feral.crit_depr end
			},
			meta = {
				order = 9, name = function() return select(2, GetItemInfo(76884)) end, type = "toggle",
                desc = L["Check if equipped"],
				set = function(info,val) options.feral.meta = not options.feral.meta; AffDots:InitSpec() end,
				get = function(info) return options.feral.meta end
			},
			sb_scan = {
				order = 10, name = L["Scan for Skull Banner"], type = "toggle",
				set = function(info,val) options.feral.sb_scan = not options.feral.sb_scan; AffDots:InitSpec() end,
				get = function(info) return options.feral.sb_scan end
			},
        }
        
        self.options.args.colors.args={
            color1 = {
                order = 1, width = "full", name = L["Current stats are better"], type = "color",
                set = function(info,r,g,b) options.color1={r,g,b}; self:SaveColors() end,
                get = function(info) return options.color1[1], options.color1[2], options.color1[3],1 end
            },
            color2 = {
                order = 2, width = "full", name = L["Current stats are same"], type = "color",
                set = function(info,r,g,b) options.color2={r,g,b}; self:SaveColors()  end,
                get = function(info) return options.color2[1], options.color2[2], options.color2[3],1 end
            },
            color3 = {
                order = 3, width = "full", name = L["Current stats are worse"], type = "color",
                set = function(info,r,g,b) options.color6={r,g,b}; self:SaveColors()  end,
                get = function(info) return options.color6[1], options.color6[2], options.color6[3],1 end
            },
            color7 = {
                order = 4, width = "full", name = L["No dot on target"], type = "color",
                set = function(info,r,g,b) options.color7={r,g,b}; self:SaveColors()  end,
                get = function(info) return options.color7[1], options.color7[2], options.color7[3],1 end
            },
            text_color = {
                order = 5, width = "full", name = L["Text color"], type = "color",
                set = function(info,r,g,b) options.color11={r,g,b} end,
                get = function(info) return options.color11[1], options.color11[2], options.color11[3],1 end
            },
            color_switch = {
                order = 6, name = L["Color switch"], type = "range",
                desc = L["Switch color only if current stats are at least x% better"],
                min = 0, max = 40, step = 1, softMin = 0, softMax=40,
                set = function(info,val) options.feral.color_switch = 1+val/100 end,
                get = function(info) return (options.feral.color_switch-1)*100 end
            },
        }
    else
        self.options.args.spec.name = "Specialization"
        self.options.args.spec.args = {
            spacer1 = {
                order = 1,type = "description",name=L["Specialization not supported, AffDots deactivated"], width="full",
            },
        }
    end
end

function AffDots:FindColor(GUID,time_left,spell,pandemic)
    if not targets[GUID..spell] or not dot_damage[spell] then
        return 7
    end
    local  current    = dot_damage[spell][1]
    local  on_target  = targets[GUID..spell][1]
    local color
    
    if on_target*colorSwitch < current then 
        color = 1
    elseif on_target > current then 
        color = 6
    else
        color = 2
    end
    
    return color
end

-- Fix for not being able to detect talent switches properly because spec is a local variable in AffDots
-- Remove as soon as possible
function AffDotsHandler(msg, editBox)
	AffDots:GetLayoutOptions()
	AffDots:GetSpecOptions("DRUID"..spec)
	LibStub("AceConfigDialog-3.0"):Open("AffDots")
end
SlashCmdList["AFFDOTS"] = AffDotsHandler
